@@ -2,6 +2,8 @@ import os
 import mysql.connector
 from mysql.connector import errorcode
 from tqdm import tqdm
+from differentiate import differentiate
+from databasetools.sqlprep import prepare_sql
 
 
 def get_column_value_strings(columns, query_type='insert'):
@@ -26,7 +28,8 @@ def get_column_value_strings(columns, query_type='insert'):
 
 
 def join_columns(cols):
-    return ", ".join([i for i in cols])
+    """Join list of columns into a string for a SQL query"""
+    return ", ".join([i for i in cols]) if isinstance(cols, list) else cols
 
 
 class MySQL:
@@ -85,7 +88,7 @@ class MySQL:
         """Commit the changes made during the current connection."""
         self._cnx.commit()
 
-    def _fetch(self, statement):
+    def _fetch(self, statement, _print=False):
         """Execute a SQL query and return values."""
         # Execute statement
         self._cursor.execute(statement)
@@ -94,8 +97,9 @@ class MySQL:
             if len(row) == 1:
                 rows.append(row[0])
             else:
-                rows.append(row)
-        self._printer('\tMySQL rows successfully queried', len(rows))
+                rows.append(list(row))
+        if _print:
+            self._printer('\tMySQL rows successfully queried', len(rows))
 
         # Return a single item if the list only has one item
         if len(rows) == 1:
@@ -111,12 +115,12 @@ class MySQL:
         self._cursor.executemany(command)
         self._commit()
 
-    def select(self, table, cols):
+    def select(self, table, cols, _print=True):
         """Query only certain columns from a table and every row."""
         # Concatenate statement
         cols_str = join_columns(cols)
         statement = ("SELECT " + cols_str + " FROM " + str(table))
-        return self._fetch(statement)
+        return self._fetch(statement, _print)
 
     def select_where(self, table, cols, where):
         """Query certain columns from a table where a particular value is found."""
@@ -171,17 +175,76 @@ class MySQL:
             self._cursor.executemany(statement, values)
             self._printer('\tMySQL rows (' + str(len(values)) + ') successfully INSERTED')
 
-    def update(self, table, columns, values, where):
-        """Update the values of a particular row where a value is met."""
-        where_col, where_val = where  # Unpack WHERE clause dictionary into tuple
-        cols = get_column_value_strings(columns, query_type='update')  # Get SET clause string
+    def insert_uniques(self, table, columns, values):
+        """
+        Insert multiple rows into a table that do not already exist.
+
+        If the rows primary key already exists, the rows values will be updated.
+        If the rows primary key does not exists, a new row will be inserted
+        """
+        # Rows that exist in the table
+        existing_rows = self.select(table, columns)
+
+        # Rows that DO NOT exist in the table
+        unique = differentiate(existing_rows, values)
+
+        # Keys that exist in the table
+        keys = self.get_primary_key_values(table)
+
+        # Primary key's column index
+        pk_col = self.get_primary_key(table)
+        pk_index = columns.index(pk_col)
+
+        # Split list of unique rows into list of rows to update and rows to insert
+        to_insert, to_update = [], []
+        for index, row in enumerate(unique):
+            # Primary key is not in list of pk values, insert new row
+            if row[pk_index] not in keys:
+                to_insert.append(unique[index])
+
+            # Primary key exists, update row rather than insert
+            elif row[pk_index] in keys:
+                to_update.append(unique[index])
+
+        # Insert new rows
+        if len(to_insert) > 0:
+            self.insert_many(table, columns, to_insert)
+
+        # Update existing rows
+        if len(to_update) > 0:
+            self.update_many(table, columns, to_update, pk_col, 0)
+
+    @staticmethod
+    def _update_statement(table, columns, where):
+        """Generate a SQL update statement."""
+        # Unpack WHERE clause dictionary into tuple
+        where_col, where_val = where
+
+        # Create column string from list of values
+        cols = get_column_value_strings(columns, query_type='update')
 
         # Concatenate statement
-        statement = ("UPDATE " + str(table) + " SET " + str(cols) + ' WHERE ' + str(where_col) + '=' + str(where_val))
+        return "UPDATE " + str(table) + " SET " + str(cols) + ' WHERE ' + str(where_col) + '=' + str(where_val)
+
+    def update(self, table, columns, values, where):
+        """
+        Update the values of a particular row where a value is met.
+
+        :param table: table name
+        :param columns: column(s) to update
+        :param values: updated values
+        :param where: tuple, (where_column, where_value)
+        """
+        statement = self._update_statement(table, columns, where)
 
         # Execute statement
         self._cursor.execute(statement, values)
-        self._printer('\tMySQL rows (' + str(len(values)) + ') successfully UPDATED')
+        self._printer('\tMySQL cols (' + str(len(values)) + ') successfully UPDATED')
+
+    def update_many(self, table, columns, values, where_col, where_index):
+        """Update the values of several rows."""
+        for row in values:
+            self.update(table, columns, row, (where_col, row[where_index]))
 
     def truncate(self, table):
         """Empty a table by deleting all of its rows."""
@@ -191,6 +254,7 @@ class MySQL:
 
     def truncate_database(self):
         """Drop all tables in a database."""
+        self.enable_printing = False
         # Loop through each table and execute a drop command
         return [self.drop_table(table) for table in
                 tqdm(self.tables, total=len(self.tables), desc='Truncating database')]
@@ -267,12 +331,12 @@ class MySQL:
         # Dump failed commands to text file
         if len(fails) > 1:
             # Re-add semi-colon separator
-            fails = [com + ';' for com in fails]
+            fails = [com + ';\n' for com in fails]
             self._printer(len(fails), 'total failed commands')
 
             # Dump failed commands to text file in the same directory as the script
             txt_file = os.path.join(os.path.dirname(sql_script),
-                                    str(os.path.basename(sql_script).rsplit('.')[0]) + '_fails.txt')
+                                    str(os.path.basename(sql_script).strip('_fails').rsplit('.')[0]) + '_fails.sql')
             self._printer('Fail commands dumped to', txt_file)
             with open(txt_file, 'w') as txt:
                 txt.writelines(fails)
@@ -284,17 +348,25 @@ class MySQL:
 
         # If with_headers is True, insert headers to first row before returning
         if with_headers:
-            f.insert(0, ('Type', 'Null', 'Key', 'Default', 'Extra'))
+            f.insert(0, ['Column', 'Type', 'Null', 'Key', 'Default', 'Extra'])
         return f
 
+    def get_primary_key(self, table):
+        """Retrieve the column which is the primary key for a table."""
+        for column in self.get_schema(table):
+            if 'pri' in column[3].lower():
+                return column[0]
+
+    def get_primary_key_values(self, table):
+        """Retrieve a list of primary key values in a table"""
+        return self.select(table, self.get_primary_key(table), _print=False)
+
     def count_rows(self, table):
         """Get the number of rows in a particular table"""
-        statement = 'SELECT COUNT(*) FROM ' + table
-        return self._fetch(statement)
+        return self.select(table, 'COUNT(*)', False)
 
     def count_rows_all(self):
         """Get the number of rows for every table in the database."""
-        self.enable_printing = False
         return {table: self.count_rows(table) for table in self.tables}
 
 
